"""
Output various aspects of topic tree to string or file.

:copyright: Copyright since 2006 by Oliver Schoenborn, all rights reserved.
:license: BSD, see LICENSE_BSD_Simple.txt for details.
"""

from typing import TextIO
from textwrap import TextWrapper

from ..core.topictreetraverser import (ITopicTreeVisitor, TopicTreeTraverser)


class TopicTreePrinter(ITopicTreeVisitor):
    """
    Example topic tree visitor that prints a prettified representation
    of topic tree by doing a depth-first traversal of topic tree and
    print information at each (topic) node of tree. Extra info to be
    printed is specified via the 'extra' kwarg. Its value must be a
    list of characters, the order determines output order:
    - D: print description of topic
    - a: print kwarg names only
    - A: print topic kwargs and their description
    - L: print listeners currently subscribed to topic

    E.g. TopicTreePrinter(extra='LaDA') would print, for each topic,
    the list of subscribed listeners, the topic's list of kwargs, the
    topic description, and the description for each kwarg,

        >>> Topic "delTopic"
           >> Listeners:
              > listener1_2880 (from yourModule)
              > listener2_3450 (from yourModule)
           >> Names of Message arguments:
              > arg1
              > arg2
           >> Description: whenever a topic is deleted
           >> Descriptions of Message arguments:
              > arg1: (required) its description
              > arg2: some other description

    """

    allowedExtras = frozenset('DAaL')  # must NOT change
    ALL_TOPICS_NAME = 'ALL_TOPICS'  # output for name of 'all topics' topic

    def __init__(self, extra=None, width: int = 70, indentStep: int = 4,
                 bulletTopic: str = '\\--', bulletTopicItem: str = '|==', bulletTopicArg: str = '-',
                 fileObj: TextIO = None):
        """
        Topic tree printer will print listeners for each topic only
        if printListeners is True. The width will be used to limit
        the width of text output, while indentStep is the number of
        spaces added each time the text is indented further. The
        three bullet parameters define the strings used for each
        item (topic, topic items, and kwargs).
        """
        self.__contentMeth = dict(
            D=self.__printTopicDescription,
            A=self.__printTopicArgsAll,
            a=self.__printTopicArgNames,
            L=self.__printTopicListeners)
        assert self.allowedExtras == set(self.__contentMeth.keys())
        import sys
        self.__destination = fileObj or sys.stdout
        self.__output = []

        self.__content = extra or ''
        unknownSel = set(self.__content) - self.allowedExtras
        if unknownSel:
            msg = 'These extra chars not known: %s' % ','.join(unknownSel)
            raise ValueError(msg)

        self.__width = width
        self.__wrapper = TextWrapper(width)
        self.__indent = 0
        self.__indentStep = indentStep
        self.__topicsBullet = bulletTopic
        self.__topicItemsBullet = bulletTopicItem
        self.__topicArgsBullet = bulletTopicArg

    def getOutput(self):
        return '\n'.join(self.__output)

    def _doneTraversal(self):
        if self.__destination is not None:
            self.__destination.write(self.getOutput())

    def _onTopic(self, topicObj):
        """This gets called for each topic. Print as per specified content."""

        # topic name
        self.__wrapper.width = self.__width
        indent = self.__indent
        if topicObj.isAll():
            topicName = self.ALL_TOPICS_NAME
        else:
            topicName = topicObj.getNodeName()
        head = '%s Topic "%s"' % (self.__topicsBullet, topicName)
        self.__output.append(self.__formatDefn(indent, head))
        indent += self.__indentStep

        # each extra content (assume constructor verified that chars are valid)
        for item in self.__content:
            function = self.__contentMeth[item]
            function(indent, topicObj)

    def _startChildren(self):
        """Increase the indent"""
        self.__indent += self.__indentStep

    def _endChildren(self):
        """Decrease the indent"""
        self.__indent -= self.__indentStep

    def __formatDefn(self, indent, item, defn='', sep=': '):
        """
        Print a definition: a block of text at a certain indent,
        has item name, and an optional definition separated from
        item by sep.
        """
        if defn:
            prefix = '%s%s%s' % (' ' * indent, item, sep)
            self.__wrapper.initial_indent = prefix
            self.__wrapper.subsequent_indent = ' ' * (indent + self.__indentStep)
            return self.__wrapper.fill(defn)
        else:
            return '%s%s' % (' ' * indent, item)

    def __printTopicDescription(self, indent, topicObj):
        # topic description
        defn = '%s Description' % self.__topicItemsBullet
        self.__output.append(
            self.__formatDefn(indent, defn, topicObj.getDescription()))

    def __printTopicArgsAll(self, indent, topicObj, desc=True):
        # topic kwargs
        args = topicObj.getArgDescriptions()
        if args:
            # required, optional, complete = topicObj.getArgs()
            headName = 'Names of Message arguments:'
            if desc:
                headName = 'Descriptions of message arguments:'
            head = '%s %s' % (self.__topicItemsBullet, headName)
            self.__output.append(self.__formatDefn(indent, head))
            tmpIndent = indent + self.__indentStep
            required = topicObj.getArgs()[0]
            for key, arg in args.items():  # iter in 3, list in 2 ok
                if not desc:
                    arg = ''
                elif key in required:
                    arg = '(required) %s' % arg
                msg = '%s %s' % (self.__topicArgsBullet, key)
                self.__output.append(self.__formatDefn(tmpIndent, msg, arg))

    def __printTopicArgNames(self, indent, topicObj):
        self.__printTopicArgsAll(indent, topicObj, False)

    def __printTopicListeners(self, indent, topicObj):
        if topicObj.hasListeners():
            item = '%s Listeners:' % self.__topicItemsBullet
            self.__output.append(self.__formatDefn(indent, item))
            tmpIndent = indent + self.__indentStep
            for listener in topicObj.getListenersIter():
                item = '%s %s (from %s)' % (self.__topicArgsBullet, listener.name(), listener.module())
                self.__output.append(self.__formatDefn(tmpIndent, item))


def printTreeDocs(rootTopic=None, topicMgr=None, **kwargs):
    """
    Print out the topic tree to a file (or file-like object like a
    StringIO), starting at rootTopic. If root topic should be root of
    whole tree, get it from pub.getDefaultTopicTreeRoot().
    The treeVisitor is an instance of pub.TopicTreeTraverser.

    Printing the tree docs would normally involve this::

        from pubsub import pub
        from pubsub.utils.topictreeprinter import TopicTreePrinter
        traverser = pub.TopicTreeTraverser( TopicTreePrinter(**kwargs) )
        traverser.traverse( pub.getDefaultTopicTreeRoot() )

    With printTreeDocs, it looks like this::

        from pubsub import pub
        from pubsub.utils import printTreeDocs
        printTreeDocs()

    The kwargs are the same as for TopicTreePrinter constructor:
    extra(None), width(70), indentStep(4), bulletTopic, bulletTopicItem,
    bulletTopicArg, fileObj(stdout). If fileObj not given, stdout is used.
    """
    if rootTopic is None:
        if topicMgr is None:
            from .. import pub
            topicMgr = pub.getDefaultTopicMgr()
        rootTopic = topicMgr.getRootAllTopics()

    printer = TopicTreePrinter(**kwargs)
    traverser = TopicTreeTraverser(printer)
    traverser.traverse(rootTopic)



"""
Top-level functionality related to message listeners.

:copyright: Copyright since 2006 by Oliver Schoenborn, all rights reserved.
:license: BSD, see LICENSE_BSD_Simple.txt for details.
"""

from types import ModuleType
from typing import Callable, Mapping, Any, Sequence

from .callables import (
    getID,
    getArgs,
    ListenerMismatchError,
    CallArgsInfo,
    AUTO_TOPIC as _AUTO_ARG,
    UserListener,
)
from .weakmethod import getWeakRef, WeakRef
from .annotations import annotationType

__all__ = [
    'Listener',
    'IListenerExcHandler',
    'ListenerValidator'
]


@annotationType
class Topic:
    pass


@annotationType
class Listener:
    """Wrapper of a UserListener"""
    pass


class IListenerExcHandler:
    """
    Interface class base class for any handler given to pub.setListenerExcHandler()
    Such handler is called whenever a listener raises an exception during a
    pub.sendMessage(). Example::

        from pubsub import pub

        class MyHandler(pub.IListenerExcHandler):
            def __call__(self, listenerID, topicObj):
                ... do something with listenerID ...

        pub.setListenerExcHandler(MyHandler())

    Without an exception handler, the sendMessage() will fail.
    """

    def __call__(self, listenerID: str, topicObj: Topic):
        raise NotImplementedError('%s must override __call__()' % self.__class__)


class Listener:
    """
    Wraps a callable (UserListener) so it can be stored by weak reference and introspected
    to verify that it adheres to a topic's MDS.

    A Listener instance has the same hash value as the callable that it wraps.

    Callables that have 'argName=pub.AUTO_TOPIC' as a kwarg will
    be given the Topic object for the message sent by sendMessage().
    Such a Listener will have wantsTopicObjOnCall() True.

    Callables that have a '** kargs' argument will receive all message data, not just that for
    the topic they are subscribed to. Such a listener will have wantsAllMessageData() True.
    """

    AUTO_TOPIC = _AUTO_ARG

    def __init__(self, callable_obj: UserListener, argsInfo: CallArgsInfo, curriedArgs: Mapping[str, Any] = None,
                 onDead: Callable[[Listener], None] = None):
        """
        Use callable_obj as a listener of topicName. The argsInfo is the
        return value from a Validator, ie an instance of callables.CallArgsInfo.
        If given, the onDead will be called with self as parameter, if/when
        callable_obj gets garbage collected (callable_obj is held only by weak
        reference).
        """
        # set call policies
        self.acceptsAllKwargs = argsInfo.acceptsAllKwargs
        self.curriedArgs = curriedArgs

        self._autoTopicArgName = argsInfo.autoTopicArgName
        self._callable = getWeakRef(callable_obj, self.__notifyOnDead)
        self.__onDead = onDead

        # save identity now in case callable dies:
        name, mod = getID(callable_obj)  #
        self.__nameID = name
        self.__module = mod
        self.__id = str(id(callable_obj))[-4:]  # only last four digits of id
        self.__hash = hash(callable_obj)

    def name(self) -> str:
        """
        Return a human readable name for listener, based on the
        listener's type name and its id (as obtained from id(listener)). If
        caller just needs name based on type info, specify instance=False.
        Note that the listener's id() was saved at construction time (since
        it may get garbage collected at any time) so the return value of
        name() is not necessarily unique if the callable has died (because
        id's can be re-used after garbage collection).
        """
        return '%s_%s' % (self.__nameID, self.__id)

    def typeName(self) -> str:
        """
        Get a type name for the listener. This is a class name or
        function name, as appropriate.
        """
        return self.__nameID

    def module(self) -> ModuleType:
        """
        Get the module in which the callable was defined.
        """
        return self.__module

    def getCallable(self) -> UserListener:
        """
        Get the listener that was given at initialization. Note that
        this could be None if it has been garbage collected (e.g. if it was
        created as a wrapper of some other callable, and not stored
        locally).
        """
        return self._callable()

    def isDead(self) -> bool:
        """Return True if this listener died (has been garbage collected)"""
        return self._callable() is None

    def wantsTopicObjOnCall(self) -> bool:
        """True if this listener wants topic object: it has a arg=pub.AUTO_TOPIC"""
        return self._autoTopicArgName is not None

    def wantsAllMessageData(self) -> bool:
        """True if this listener wants all message data: it has a ** kwargs argument"""
        return self.acceptsAllKwargs

    def setCurriedArgs(self, **curriedArgs):
        """
        Curry the wrapped listener so it appears to *not* have list(curriedArgs) among its parameters.
        The curriedArgs key-value pairs will be given to wrapped listener at call time.
        """
        if curriedArgs.keys() != self.curriedArgs.keys():
            raise ValueError(
                "Listener '{}' already subscribed with a different set of pure curried args ({} != {})"
                .format(self, curriedArgs.keys(), self.curriedArgs.keys()))

        self.curriedArgs = curriedArgs

    def _unlinkFromTopic_(self):
        """Tell self that it is no longer used by a Topic. This allows to break some cyclical references."""
        self.__onDead = None

    def _calledWhenDead(self):
        raise RuntimeError('BUG: Dead Listener called, still subscribed!')

    def __notifyOnDead(self, _: WeakRef):
        """This gets called when listener weak ref has died. Propagate info to Topic."""
        notifyDeath = self.__onDead
        self._unlinkFromTopic_()
        if notifyDeath is not None:
            notifyDeath(self)

    def __eq__(self, rhs: Listener):
        """
        Compare for equality to rhs. This returns true if rhs has our id id(rhs) is same as
        id(self) or id(callable in self).
        """
        if id(self) == id(rhs):
            return True

        c1 = self._callable()
        try:
            c2 = rhs._callable()
        except Exception:
            # then rhs is not a Listener, compare with c1
            return c1 == rhs

        # both side of == are Listener, but always compare unequal if both dead
        if c2 is None and c1 is None:
            return False

        return c1 == c2

    def __ne__(self, rhs: Listener):
        """Counterpart to __eq__ MUST be defined... equivalent to 'not (self == rhs)'."""
        return not self.__eq__(rhs)

    def __hash__(self):
        """
        Hash is an optimization for dict/set searches, it need not return different numbers for every different object.
        """
        return self.__hash

    def __str__(self):
        """String rep is the callable"""
        return self.__nameID

    def __call__(self, kwargs: Mapping[str, Any], actualTopic: Topic, allKwargs: Mapping[str, Any] = None):
        """
        Call the listener with **kwargs. Note that it raises RuntimeError
        if listener is dead. Should always return True (False would require
        the callable_obj be dead but self hasn't yet been notified of it...).
        """
        if self.acceptsAllKwargs:
            kwargs = allKwargs or kwargs  # if allKwargs is None then use kwargs

        orig_kwargs = kwargs

        # combine with curried args; Note: this overrides topic arg if present:
        if self.curriedArgs:
            if kwargs:
                kwargs = kwargs.copy()
                kwargs.update(self.curriedArgs)
            else:
                kwargs = self.curriedArgs

        if self._autoTopicArgName is not None:
            if kwargs is orig_kwargs:
                kwargs = kwargs.copy()
            kwargs[self._autoTopicArgName] = actualTopic

        # call:
        cb = self._callable()
        if cb is None:
            self._calledWhenDead()
        cb(**kwargs)

        return True


class ListenerValidator:
    """
    Validates listeners. It checks whether the listener given to
    validate() method complies with required and optional arguments
    specified for topic.

    Do not accept any required args or *args; accept any **kwarg,
    and require that the Listener have at least all the kwargs (can
    have extra) of Topic.
    """

    def __init__(self, topicArgs: Sequence[str], topicKwargs: Sequence[str]):
        """
        :param topicArgs: list of argument names that will be required when sending
            a message to listener. Hence order of items in topicArgs matters.
        :param topicKwargs: list of argument names that will be optional, ie given as keyword arguments
            when sending a message to listener. The list is unordered. """
        self._topicArgs = set(topicArgs)
        self._topicKwargs = set(topicKwargs)

    def validate(self, listener: UserListener, curriedArgNames: Sequence[str] = None) -> CallArgsInfo:
        """
        Validate that listener (with, optionally, given curried parameters) satisfies the requirements of
        being a topic listener.

        :param listener: the callable to validate
        :param curriedArgNames: the list of parameter names to treat as curried
        :returns: a CallArgsInfo object containing information about the listener's call arguments, such as
            whether listener wants topic name (signified by a kwarg value = AUTO_TOPIC in listener signature).
        :raises ListenerMismatchError: if listener not usable for topic
        """
        paramsInfo = getArgs(listener)
        self.__validateArgs(listener, paramsInfo, curriedArgNames)

        return paramsInfo

    # noinspection PyIncorrectDocstring
    def isValid(self, listener: UserListener, curriedArgNames: Sequence[str] = None) -> bool:
        """Same as validate() but returns True/False instead of raising an exception."""
        try:
            self.validate(listener, curriedArgNames=curriedArgNames)
            return True
        except ListenerMismatchError:
            return False

    def __validateArgs(self, listener: UserListener, paramsInfo: CallArgsInfo, curriedArgNames: Sequence[str]):
        # accept **kwargs
        # accept *args

        # check if listener missing params (only possible if
        # paramsInfo.acceptsAllKwargs is False)
        if not paramsInfo.acceptsAllKwargs:
            allTopicMsgArgs = self._topicArgs | self._topicKwargs
            allParams = set(paramsInfo.allParams)
            missingParams = allTopicMsgArgs - allParams
            if missingParams:
                msg = 'needs to accept %s more args (%s)' \
                      % (len(missingParams), ', '.join(missingParams))
                raise ListenerMismatchError(msg, listener, missingParams)
        else:
            # then can accept that some parameters missing from listener
            # signature
            pass

        if curriedArgNames:
            unrecognizedCurried = set(curriedArgNames).difference(paramsInfo.allParams)
            if unrecognizedCurried:
                msg = 'does not have following args: (%s)' % ', '.join(unrecognizedCurried)
                raise ListenerMismatchError(msg, listener, unrecognizedCurried)

            curriedTopicArgs = set(curriedArgNames).intersection(self._topicArgs | self._topicKwargs)
            if curriedTopicArgs:
                msg = 'curried args (%s) are topic args, not allowed' % ', '.join(curriedTopicArgs)
                raise ListenerMismatchError(msg, listener, curriedTopicArgs)

        # check if there are extra required parameters in listener signature:
        extraArgs = set(paramsInfo.getRequiredArgs()) - self._topicArgs
        if extraArgs and curriedArgNames:
            extraArgs = extraArgs.difference(curriedArgNames)
        if extraArgs:
            msg = 'required args (%s) not allowed (could curry them), ' % ','.join(extraArgs)
            if self._topicArgs:
                msg += 'topic req\'d args are (%s)' % ', '.join(self._topicArgs)
            else:
                msg += 'topic has no required args'

            # now make sure listener doesn't require params that are optional in TMS:
            missingDefaultVals = extraArgs.intersection(self._topicKwargs)
            if missingDefaultVals:
                msg += ' (params (%s) are req\'d in listener, optional in topic )' % ', '.join(missingDefaultVals)

            raise ListenerMismatchError(msg, listener, extraArgs)



"""
Various utilities used by topic-related modules.

:copyright: Copyright since 2006 by Oliver Schoenborn, all rights reserved.
:license: BSD, see LICENSE_BSD_Simple.txt for details.

"""

from textwrap import TextWrapper, dedent
import sys
from typing import Tuple, List, Sequence, Mapping, Dict, Callable, Any, Optional, Union, TextIO
from .topicexc import TopicNameError

__all__ = []

UNDERSCORE = '_'  # topic name can't start with this
# just want something unlikely to clash with user's topic names
ALL_TOPICS = 'ALL_TOPICS'


class WeakNone:
    """
    Pretend to be a weak reference to nothing. Used by ArgsInfos to
    refer to parent when None so no if-else blocks needed.
    """

    def __call__(self):
        return None


def smartDedent(paragraph: str) -> str:
    """
    Dedent paragraph using textwrap.dedent(), but properly dedents
    even if the first line of paragraph does not contain blanks.
    This handles the case where a user types a documentation string as
        '''A long string spanning
        several lines.'''
    """
    if paragraph.startswith(' '):
        para = dedent(paragraph)
    else:
        lines = paragraph.split('\n')
        exceptFirst = dedent('\n'.join(lines[1:]))
        para = lines[0] + exceptFirst

    return para


import re

_validNameRE = re.compile(r'[-0-9a-zA-Z]\w*')


def validateName(topicName: str):
    """Raise TopicNameError if nameTuple not valid as topic name."""
    topicNameTuple = tupleize(topicName)
    if not topicNameTuple:
        reason = 'name tuple must have at least one item!'
        raise TopicNameError(None, reason)

    class topic:
        pass

    for subname in topicNameTuple:
        if not subname:
            reason = 'can\'t contain empty string or None'
            raise TopicNameError(topicNameTuple, reason)

        if subname.startswith(UNDERSCORE):
            reason = 'must not start with "%s"' % UNDERSCORE
            raise TopicNameError(topicNameTuple, reason)

        if subname == ALL_TOPICS:
            reason = 'string "%s" is reserved for root topic' % ALL_TOPICS
            raise TopicNameError(topicNameTuple, reason)

        if _validNameRE.match(subname) is None:
            reason = 'element #%s ("%s") has invalid characters' % \
                     (1 + list(topicNameTuple).index(subname), subname)
            raise TopicNameError(topicNameTuple, reason)


def stringize(topicName: Sequence[str]) -> str:
    """
    If topicName is a string, just return it
    as is. If it is a topic definition object (ie an object that has
    'msgDataSpec' as data member), return the dotted name of corresponding
    topic. Otherwise, assume topicName is a tuple and convert it to to a
    dotted name i.e. ('a','b','c') => 'a.b.c'. Empty name is not allowed
    (ValueError). The reverse operation is tupleize(topicName).
    """
    if isinstance(topicName, str):
        return topicName

    if hasattr(topicName, "_topicNameStr"):
        return topicName._topicNameStr

    try:
        name = '.'.join(topicName)
    except Exception:
        exc = sys.exc_info()[1]
        raise TopicNameError(topicName, str(exc))

    return name


def tupleize(topicName: str) -> Tuple[str, ...]:
    """
    If topicName is a tuple of strings, just return it as is. Otherwise,
    convert it to tuple, assuming dotted notation used for topicName. I.e.
    'a.b.c' => ('a','b','c'). Empty topicName is not allowed (ValueError).
    The reverse operation is stringize(topicNameTuple).
    """
    # assume name is most often str; if more often tuple,
    # then better use isinstance(name, tuple)
    if hasattr(topicName, "msgDataSpec"):
        topicName = topicName._topicNameStr
    if isinstance(topicName, str):
        topicTuple = tuple(topicName.split('.'))
    else:
        topicTuple = tuple(topicName)  # assume already tuple of strings

    if not topicTuple:
        raise TopicNameError(topicTuple, "Topic name can't be empty!")

    return topicTuple



"""
:copyright: Copyright since 2006 by Oliver Schoenborn, all rights reserved.
:license: BSD, see LICENSE_BSD_Simple.txt for details.
"""

from typing import Tuple, List, Sequence, Mapping, Dict, Callable, Any, Optional, Union

from .topicmgr import (
    TopicManager,
    TreeConfig
)

from .listener import IListenerExcHandler, Listener, UserListener
from .notificationmgr import INotificationHandler

TopicFilter = Callable[[str], bool]
ListenerFilter = Callable[[Listener], bool]


class Publisher:
    """
    Represent the class that send messages to listeners of given
    topics and that knows how to subscribe/unsubscribe listeners
    from topics.
    """

    def __init__(self, treeConfig: TreeConfig = None):
        """
        :param treeConfig: the TreeConfig instance to use; if None, a new one is created
        """
        self.__treeConfig = treeConfig or TreeConfig()
        self.__topicMgr = TopicManager(self.__treeConfig)

    def getTopicMgr(self) -> TopicManager:
        """Get the topic manager created for this publisher."""
        return self.__topicMgr

    def getListenerExcHandler(self) -> IListenerExcHandler:
        """
        Get the listener exception handler that was registered
        via setListenerExcHandler(), or None of none registered.
        """
        return self.__treeConfig.listenerExcHandler

    def setListenerExcHandler(self, handler: IListenerExcHandler):
        """Set the function to call when a listener raises an exception during a sendMessage()."""
        self.__treeConfig.listenerExcHandler = handler

    def addNotificationHandler(self, handler: INotificationHandler):
        """Add a handler for tracing pubsub activity."""
        self.__treeConfig.notificationMgr.addHandler(handler)

    def clearNotificationHandlers(self):
        """
        Remove all notification handlers that were added via
        self.addNotificationHandler().
        """
        self.__treeConfig.notificationMgr.clearHandlers()

    def setNotificationFlags(self, **kwargs: Mapping[str, Optional[bool]]):
        """
        Set the notification flags on or off for each type of
        pubsub activity. The kwargs keys can be any of the following:

        - subscribe:    if True, get notified whenever a listener subscribes to a topic;
        - unsubscribe:  if True, get notified whenever a listener unsubscribes from a topic;
        - deadListener: if True, get notified whenever a subscribed listener has been garbage-collected;
        - sendMessage:  if True, get notified whenever sendMessage() is called;
        - newTopic:     if True, get notified whenever a new topic is created;
        - delTopic:     if True, get notified whenever a topic is "deleted" from topic tree;
        - all:          set all of the above to the given value (True or False).

        The kwargs that are not given or None are left at their current value. Those that are
        False will cause corresponding notification to be silenced. The 'all'
        is set first, then the others. E.g.

            mgr.setFlagStates(all=True, delTopic=False)

        will toggle all notifications on, but will turn off the 'delTopic'
        notification.
        """
        self.__treeConfig.notificationMgr.setFlagStates(**kwargs)

    def getNotificationFlags(self) -> Mapping[str, bool]:
        """Return a dictionary with the notification flag states."""
        return self.__treeConfig.notificationMgr.getFlagStates()

    def setTopicUnspecifiedFatal(self, newVal: bool = True, checkExisting: bool = True) -> bool:
        """
        Changes the creation policy for topics.

        By default, pubsub will accept topic names for topics that
        don't have a message data specification (MDS). This default behavior
        makes pubsub easier to use initially, but allows topic
        names with typos to go uncaught in common operations such as
        sendMessage() and subscribe(). In a large application, this
        can lead to nasty bugs. Pubsub's default behavior is equivalent
        to setTopicUnspecifiedFatal(false).

        When called with newVal=True, any future pubsub operation that
        requires a topic (such as subscribe and sendMessage) will require
        an MDS; if none is available, pubsub will raise a TopicDefnError
        exception.

        If checkExisting is not given or True, all existing
        topics are validated. A TopicDefnError exception is
        raised if one is found to be incomplete (has hasMDS() false).

        Returns previous value of newVal.

        Note that this method can be used in several ways:

        1. Only use it in your application when something is not working
           as expected: just add a call at the beginning of your app when
           you have a problem with topic messages not being received
           (for instance), and remove it when you have fixed the problem.

        2. Use it from the beginning of your app and never use newVal=False:
           add a call at the beginning of your app and you leave it in
           (forever), and use Topic Definition Providers to provide the
           listener specifications. These are easy to use via the
           pub.addTopicDefnProvider().

        3. Use it as in #1 during app development, and once stable, use
           #2. This is easiest to do in combination with
           pub.exportTopicTreeSpec().
        """
        oldVal = self.__treeConfig.raiseOnTopicUnspecified
        self.__treeConfig.raiseOnTopicUnspecified = newVal

        if newVal and checkExisting:
            self.__topicMgr.checkAllTopicsHaveMDS()

        return oldVal

    def subscribe(self, listener: UserListener, topicName: str, **curriedArgs) -> Listener:
        """
        Subscribe listener to named topic. Raises ListenerMismatchError
        if listener isn't compatible with the topic's MDS. Returns
        (pubsub.core.Listener, success), where success is False if listener
        was already subscribed. The pub.core.Listener wraps the callable
        subscribed and provides introspection-based info about
        the callable. Extra keyword arguments are treated as currying of
        listener arguments.

        Example:
            pub.subscribe(listener1, 'some_topic')
            pub.subscribe(listener2, 'some_other_topic', a=2, b=3)

        In the second example, the listener2 will always receive a=2 and b=3 and
        pubsub treats it as though a and b were curried, i.e. as if the actual
        listener subscribed were a callable that did not have a or b parameters.
        Hence if some_other_topic has a or b as message data, subscription will
        raise a ListenerInadequate error.

        Note that if 'subscribe' notification is on, the handler's
        'notifySubscribe' method is called after subscription.
        """
        topicObj = self.__topicMgr.getOrCreateTopic(topicName)
        subscribedListener, success = topicObj.subscribe(listener, **curriedArgs)
        return subscribedListener, success

    def unsubscribe(self, listener: UserListener, topicName: str):
        """
        Unsubscribe from given topic. Returns the pubsub.core.Listener
        instance that was used to wrap listener at subscription
        time. Raises an TopicNameError if topicName doesn't exist.

        Note that if 'unsubscribe' notification is on, the handler's
        notifyUnsubscribe() method will be called after unsubscribing.
        """
        topicObj = self.__topicMgr.getTopic(topicName)
        unsubdLisnr = topicObj.unsubscribe(listener)

        return unsubdLisnr

    def unsubAll(self, topicName: str = None, listenerFilter: ListenerFilter = None,
                 topicFilter: Union[str, TopicFilter] = None) -> List[Listener]:
        """
        Unsubscribe all listeners of a topic.

        :param topicName: if none given, unsub from all topics.
        :param listenerFilter: filter function to apply to listeners, unsubscribe only the listeners
            that satisfy listenerFilter(listener: Listener) == True
        :param topicFilter: topic name, or a filter function to apply to topics; in latter case, only
            topics that satisfy topicFilter(topic name) == True will be affected
        :returns: list of all listeners (instances of pub.Listener) that were unsubscribed from the topic tree

        Note: this method will generate one 'unsubcribe' notification message
        (see pub.setNotificationFlags()) for each listener unsubscribed.
        """
        unsubdListeners = []

        if topicName is None:
            # unsubscribe all listeners from all topics
            topicsMap = self.__topicMgr._topicsMap
            for topicName, topicObj in topicsMap.items():
                if topicFilter is None or topicFilter(topicName):
                    tmp = topicObj.unsubscribeAllListeners(listenerFilter)
                    unsubdListeners.extend(tmp)

        else:
            topicObj = self.__topicMgr.getTopic(topicName)
            unsubdListeners = topicObj.unsubscribeAllListeners(listenerFilter)

        return unsubdListeners

    def sendMessage(self, topicName: str, **msgData):
        """
        Send a message.
        :param topicName: name of message topic (dotted or tuple format)
        :param msgData: message data (must satisfy the topic's MDS)
        """
        topicMgr = self.getTopicMgr()
        topicObj = topicMgr.getOrCreateTopic(topicName)
        topicObj.publish(**msgData)



"""
Core package of pubsub, holding the publisher, listener, and topic
object modules.

:copyright: Copyright since 2006 by Oliver Schoenborn, all rights reserved.
:license: BSD, see LICENSE_BSD_Simple.txt for details.

"""

from .publisher import Publisher

from .callables import (
    AUTO_TOPIC,
)

from .listener import (
    getID as getListenerID,
    ListenerMismatchError,
    IListenerExcHandler,
    Listener,
)
from .topicobj import (
    Topic,
    SenderUnknownMsgDataError,
    SenderMissingReqdMsgDataError,
    MessageDataSpecError,
    TopicDefnError,
    ExcHandlerError,
)

from .topicmgr import (
    TopicManager,
    TopicDefnError,
    TopicNameError,
    ALL_TOPICS,
)

from .topicdefnprovider import (
    ITopicDefnProvider,
    TopicDefnProvider,
    ITopicDefnDeserializer,
    UnrecognizedSourceFormatError,

    exportTopicTreeSpec,
    TOPIC_TREE_FROM_MODULE,
    TOPIC_TREE_FROM_STRING,
    TOPIC_TREE_FROM_CLASS,
)

from .topictreetraverser import (
    TopicTreeTraverser,
    TreeTraversal,
)

from .notificationmgr import (
    INotificationHandler,
)



#!python
# -----------------------------
# Name:     PyPubSubMonitor
# Purpose:  A pubsub activity monitor
#
# Author:   Josh English
#
# Date:     20-Aug-2008
# Last Rev: 127 (Last revision this code was checked with)
# -------------------------------------

# NOTE: The Tool and Mixin components (including the
# docstrings) borrow heavily on Robin Dunn's wxInspectionTool

### Next problem: TreeCtrl doesn't send messages...
import wx
from pubsub import pub
import pubsub.utils as utils
from pubsub import pubsubconf as conf

conf.setNotifierClass(utils.NotifyByPubsubMessage)
pub.setNotificationFlags(all=True)


### MonitorTopics
###  Three topics for Pypubsub messages to control the monitor
class MonitorTopics(utils.TopicTreeDefnSimple):
    class monitor:
        """Messages controlling the monitor"""

        class show:
            """Shows or hides the monitor"""
            show = "Boolean value to show (True) or hide (False) the monitor"
            _required = ('show',)

        class hide:
            """Hides the monitor. Same as message ('monitor.show', show=False)"""

        class toggle:
            """Toggles the monitor to be shown or hidden"""


pub.addTopicDefnProvider(MonitorTopics())


class MonitorTool:
    """
    The MonitorTool is a singleton based on the
    wx.lib.inspection.InspectionTool.
    """
    __shared_state = {}

    def __init__(self):
        self.__dict__ = self.__shared_state
        if not hasattr(self, 'initialized'):
            self.initialized = False
        pub.subscribe(self.Show, 'monitor.show')
        print
        "Made Monitor Tool"
        tobj = pub.getTopic('monitor.show')
        print
        tobj

    def Init(self, app=None):
        self._frame = None
        self._app = app
        if not self._app:
            self._app = wx.GetApp()
        self.initialized = True
        print
        "Monitor Tool Init"

    def Show(self, show):
        print
        "Monitor Tool Show"
        if not self.initialized:
            self.Init()

        parent = self._app.GetTopWindow()
        if not self._frame:
            self._frame = MonitorFrame(parent=parent)
        self._frame.Show(show)


class MonitorMixin(object):
    """
    This class is intended to be used as a mix-in with the wx.App
    class.  When used it will add the ability to popup a
    MonitorFrame window.  The default key sequence to activate the inspector is
    Ctrl-Alt-M (or Cmd-Alt-M on Mac) but this can be changed via
    parameters to the `Init` method, or the application can call
    `ShowInspectionTool` from other event handlers if desired.

    To use this class simply derive a class from wx.App and
    MonitorMixin and then call the `Init` method from the app's
    OnInit.
    """

    def Init(self, alt=True, cmd=True, shift=False, keycode=ord('M')):
        self.Bind(wx.EVT_KEY_DOWN, self._OnKeyPress)
        self._alt = alt
        self._cmd = cmd
        self._shift = shift
        self._keycode = keycode
        MonitorTool().Init(self)
        print
        "Mixin Init"

    def _OnKeyPress(self, evt):
        if evt.AltDown() == self._alt and \
                        evt.CmdDown() == self._cmd and \
                        evt.ShiftDown() == self._shift and \
                        evt.GetKeyCode() == self._keycode:
            self.ShowMonitorTool()
        else:
            evt.Skip()

    def ShowMonitorTool(self):
        print
        "Mixin Show"
        MonitorTool().Show(True)


class MonitorFrame(wx.Frame):
    """MonitorFrame
    This is the main monitor frame. Interaction between varying panels
    is done through direct method calls to prevent message clutter
    """

    def __init__(self, parent):
        wx.Frame.__init__(self, parent, size=(700, 600), title="PyPubSubMonitor")
        panel = wx.Panel(self)
        splitter = wx.SplitterWindow(panel, style=wx.SP_BORDER)

        self.NotifyControl = NotifierPanel(panel)
        self.TopicTree = TopicTreePanel(splitter)
        self.Log = LogPanel(splitter)
        self.Entry = LastLogEntryPanel(panel)
        self.Count = CountPanel(panel)

        ##        redirect = Redirector(self.Log.getText())
        self.Logger = MonitorLogger(pub, self.Log)

        pub.subscribe(self.psUpdate, pub.ALL_TOPICS)
        pub.subscribe(self.psShow, 'monitor.show')
        pub.subscribe(self.psHide, 'monitor.hide')
        pub.subscribe(self.psToggle, 'monitor.toggle')

        sizer = wx.BoxSizer(wx.VERTICAL)
        row = wx.BoxSizer(wx.HORIZONTAL)
        row.Add(self.NotifyControl, 0, wx.EXPAND | wx.ALL, 3)

        splitter.SetMinimumPaneSize(20)
        splitter.SetSashGravity(0.5)
        splitter.SplitVertically(self.TopicTree, self.Log, 0)

        row.Add(splitter, 1, wx.EXPAND | wx.GROW | wx.ALL, 3)
        sizer.Add(row, 1, wx.EXPAND | wx.GROW | wx.ALL, 3)
        sizer.Add(self.Entry, 0, wx.EXPAND | wx.ALL, 3)
        sizer.Add(self.Count, 0, wx.EXPAND | wx.ALL, 3)
        panel.SetSizerAndFit(sizer)
        self.Layout()
        self.Refresh()
        self.Bind(wx.EVT_CLOSE, self.OnClose)
        print
        "Frame Init"

    def OnClose(self, evt):
        """Turn off notifications before closing down
        """
        pub.setNotificationFlags(all=False)
        evt.Skip()

    def psShow(self, show): self.Show(bool(show))

    def psHide(self): self.Hide()

    def psToggle(self): self.Show(not self.IsShown())

    def psUpdate(self, msgTopic=pub.AUTO_TOPIC):
        self.Entry.SetMessage(self.Logger.lastMsg.strip())
        self.Entry.SetTopic(self.Logger.lastTopic)
        self.Entry.SetPSTopic(self.Logger.lastLogTopic.getName())
        self.Count.count(self.Logger.lastLogTopic.getName())


class NotifierPanel(wx.Panel):
    """NotifierPanel
    Seven toggle buttons to control which kinds of messages the
    monitor will pay attention to.
    """

    def __init__(self, parent):
        wx.Panel.__init__(self, parent, name="notifier control")
        box = wx.StaticBox(self, label="Notifiers")

        allBtn = wx.ToggleButton(self, label="all")
        allBtn.SetValue(True)
        allBtn.Bind(wx.EVT_TOGGLEBUTTON, self.OnNotifyAll)

        self.subBtn = wx.ToggleButton(self, label="subscribe")
        self.subBtn.SetValue(True)
        self.subBtn.Bind(wx.EVT_TOGGLEBUTTON, self.OnSubscribe)

        self.unsubBtn = wx.ToggleButton(self, label="unsubscribe")
        self.unsubBtn.SetValue(True)
        self.unsubBtn.Bind(wx.EVT_TOGGLEBUTTON, self.OnUnsubscribe)

        self.sendBtn = wx.ToggleButton(self, label="sendMessage")
        self.sendBtn.SetValue(True)
        self.sendBtn.Bind(wx.EVT_TOGGLEBUTTON, self.OnSendMessage)

        self.newTBtn = wx.ToggleButton(self, label="newTopic")
        self.newTBtn.SetValue(True)
        self.newTBtn.Bind(wx.EVT_TOGGLEBUTTON, self.OnNewTopic)

        self.delTBtn = wx.ToggleButton(self, label="delTopic")
        self.delTBtn.SetValue(True)
        self.delTBtn.Bind(wx.EVT_TOGGLEBUTTON, self.OnDelTopic)

        self.deadBtn = wx.ToggleButton(self, label="deadListener")
        self.deadBtn.SetValue(True)
        self.deadBtn.Bind(wx.EVT_TOGGLEBUTTON, self.OnDead)

        btnpad = wx.LEFT | wx.RIGHT | wx.EXPAND
        sizer = wx.StaticBoxSizer(box, wx.VERTICAL)
        sizer.Add(allBtn, 0, btnpad, 3)
        sizer.Add(wx.StaticLine(self, wx.HORIZONTAL), 0, wx.ALL | wx.EXPAND, 2)
        sizer.Add(self.subBtn, 0, btnpad, 3)
        sizer.Add(self.unsubBtn, 0, btnpad, 3)
        sizer.Add(self.sendBtn, 0, btnpad, 3)
        sizer.Add(wx.StaticLine(self, wx.HORIZONTAL), 0, wx.ALL | wx.EXPAND, 2)
        sizer.Add(self.newTBtn, 0, btnpad, 3)
        sizer.Add(self.delTBtn, 0, btnpad, 3)
        sizer.Add(self.deadBtn, 0, btnpad, 3)

        self.SetSizerAndFit(sizer)
        self.Layout()

    def OnNotifyAll(self, evt):
        val = evt.IsChecked()
        self.subBtn.SetValue(val)
        self.unsubBtn.SetValue(val)
        self.sendBtn.SetValue(val)
        self.newTBtn.SetValue(val)
        self.delTBtn.SetValue(val)
        self.deadBtn.SetValue(val)
        pub.setNotificationFlags(all=val)

    def OnSubscribe(self, evt):
        pub.setNotificationFlags(subscribe=evt.IsChecked())

    def OnUnsubscribe(self, evt):
        pub.setNotificationFlags(unsubscribe=evt.IsChecked())

    def OnSendMessage(self, evt):
        pub.setNotificationFlags(sendMessage=evt.IsChecked())

    def OnNewTopic(self, evt):
        pub.setNotificationFlags(newTopic=evt.IsChecked())

    def OnDelTopic(self, evt):
        pub.setNotificationFlags(delTopic=evt.IsChecked())

    def OnDead(self, evt):
        pub.setNotificationFlags(deadListener=evt.IsChecked())


class TopicTreePanel(wx.Panel):
    """MonitorTreePanel
    Contains the topic tree and a few controls, including
    an instance of TopicDescriptionPanel
    """

    def __init__(self, parent, *args, **kwargs):
        wx.Panel.__init__(self, parent, *args, **kwargs)
        box = wx.StaticBox(self, label="Topic Tree")
        sizer = wx.StaticBoxSizer(box, wx.VERTICAL)
        label = wx.StaticText(self, label="RightClick on any item in the tree")
        ### Create a dummy panel
        self.tdp = wx.Panel(self)
        tree = TopicTree(self)
        refreshButton = wx.Button(self, label="Refresh")
        refreshButton.Bind(wx.EVT_BUTTON, tree._load)

        sizer.Add(label)
        sizer.Add(tree, 1, wx.EXPAND)
        sizer.Add(refreshButton, 0, wx.EXPAND | wx.ALL, 3)
        sizer.Add(self.tdp, 0, wx.EXPAND)

        self.SetSizerAndFit(sizer)
        self.Layout()

    def SetTopicDescription(self, tObj):
        """Replace the current topic description with a new one"""
        tdp = TopicDescriptionPanel(self, tObj)
        sizer = self.GetSizer()
        sizer.Detach(self.tdp)
        self.tdp.Destroy()
        sizer.Add(tdp, 0, wx.EXPAND | wx.ALL, 3)
        self.tdp = tdp
        self.Layout()
        self.Refresh()


class TopicTreeFiller(utils.ITopicTreeTraverser):
    """Link a pubsub.utils.ITopicTreeTraverser to a wxTreeCtrl
    Call traverse(topicObj) to fill
    """

    def __init__(self, wxTree=None, startTopic=None):
        if wxTree is None:
            raise RunTimeError, "TopicTreeFiller needs a wx.TreeCtrl to work with"

        self.Tree = wxTree
        if startTopic is None:
            self.startTopic = pub.getTopic(pub.ALL_TOPICS)
        self.root = self.Tree.AddRoot(str(self.startTopic.getName()))
        self.lastT = self.root
        self.Topics = [self.root]

    def _onTopic(self, topicObj):
        if topicObj is not self.startTopic:
            top = self.Tree.AppendItem(self.Topics[-1], str(topicObj.getName()))
            self.lastT = top

    def _startChildren(self):
        self.Topics.append(self.lastT)

    def _endChildren(self):
        self.Topics.pop()


class TopicTree(wx.TreeCtrl):
    """Subclass of wxTreeCtrl, sits in a TopicTreePanel object
    Uses direct method calls, even though this is exactly what
    pubsub is supposed to do. I don't want to create messasge static
    in the monitor
    """

    def __init__(self, parent):
        wx.TreeCtrl.__init__(self, parent)
        self._load()
        self.Bind(wx.EVT_TREE_SEL_CHANGED, self.OnSelection)
        self.Bind(wx.EVT_TREE_ITEM_MENU, self.OnRightClick)

    def SendPubSubMessage(self, evt):
        tName = self.GetItemText(self.GetSelection())
        tObj = pub.getTopic(str(tName))
        dlg = SendMessageDialog(self, tObj)
        dlg.ShowModal()

    def ShowListeners(self, evt):
        tName = self.GetItemText(self.GetSelection())
        tObj = pub.getTopic(str(tName))
        if tObj.hasListeners():
            llist = '\n'.join([repr(l.getCallable()) for l in tObj.getListenersIter()])
        else:
            llist = "No listeners"
        wx.MessageBox(llist, "Topic Listeners", style=wx.OK)

    def OnRightClick(self, evt):
        self.SelectItem(evt.GetItem())
        if not hasattr(self, 'sendmessageID'):
            self.sendmessageID = wx.NewId()
            self.showlistenersID = wx.NewId()
            self.Bind(wx.EVT_MENU, self.SendPubSubMessage)
            self.Bind(wx.EVT_MENU, self.ShowListeners)
        menu = wx.Menu()
        menu.Append(self.sendmessageID, "Send Message")
        menu.Append(self.showlistenersID, "Show Listeners")
        self.PopupMenu(menu)

    def OnSelection(self, evt):
        tName = self.GetItemText(self.GetSelection())
        tObj = pub.getTopic(str(tName))
        self.Parent.SetTopicDescription(tObj)

    def _load(self, evt=None):
        self.DeleteAllItems()
        trv = TopicTreeFiller(self)
        trv.traverse(pub.getTopic(pub.ALL_TOPICS))
        self.Expand(self.GetRootItem())


class SendMessageDialog(wx.Dialog):
    """SendMessageDialog
    create a dialog with the topic description panel and
    a check and send and cancel options.
    Currently can only check the arguments, and even that
    does not work (as of Pypubsub rev 126) because topic.checkArgs does
    not return the boolean value the docs claim it returns.
    To get OnSendMessage to work, there needs to be a better
    way to know what the listener expects for each argument.
    """

    def __init__(self, parent, tObj):
        wx.Dialog.__init__(self, parent, title="Send Message")
        self.topicObj = tObj
        s = wx.BoxSizer(wx.VERTICAL)
        self.tdp = TopicDescriptionPanel(self, tObj, useTextCtrl=True)
        checkButton = wx.Button(self, label="Check")
        checkButton.Bind(wx.EVT_BUTTON, self.CheckTopic)

        sendButton = wx.Button(self, id=wx.ID_OK, label="Send")
        sendButton.Bind(wx.EVT_BUTTON, self.OnSendMessage)

        cancelButton = wx.Button(self, wx.ID_CANCEL)

        s.Add(self.tdp, 1, wx.EXPAND | wx.ALL, 3)
        row = wx.BoxSizer(wx.HORIZONTAL)
        row.Add(checkButton, 0)
        row.Add(sendButton, 0, wx.LEFT, 6)
        row.Add(cancelButton, 0, wx.LEFT, 6)
        s.Add(row, 0, wx.ALIGN_CENTER | wx.TOP | wx.BOTTOM, 6)
        self.SetSizerAndFit(s)
        self.Layout()

    def CheckTopic(self, evt):
        args = {}
        for key in self.tdp.topicArgs:
            args[key] = self.tdp.topicArgs[key].GetValue()
        ##        print args
        ok = self.topicObj.checkArgs(args)
        ### currently gets None from topic.checkArgs
        ###  topic.checkArgs should return True or False
        if ok:
            wx.MessageBox('Arguments check out', 'Check Args')
        else:
            wx.MessageBox("Arguments don't check", "Check Args")

    def OnSendMessage(self, evt):
        """OnSendMessage
        This should create the dialog box,
        """
        wx.MessageBox("This feature is not complete. See the doc string", "Not Implemented")


class TopicDescriptionPanel(wx.Panel):
    """TopicDescriptionPanel(parent, tObj)
    Creates a panel describing the topic object
    """

    def __init__(self, parent, tObj, useTextCtrl=False):
        wx.Panel.__init__(self, parent, name="topicDescription")
        box = wx.StaticBox(self, label="Topic Information")
        sizer = wx.StaticBoxSizer(box, wx.VERTICAL)
        dtext = "%s [%s]\n(%d listener%s)" % \
                (tObj.getName(),
                 tObj.getDescription(),
                 tObj.getNumListeners(),
                 "" if tObj.getNumListeners() == 1 else "s")
        self.desc = wx.StaticText(self, label=dtext)
        self.desc.Wrap(self.Parent.GetSize()[0] - 6)
        sizer.Add(self.desc, 0, wx.EXPAND | wx.ALL, 2)
        grid = wx.FlexGridSizer(cols=2, vgap=3, hgap=2)
        grid.AddGrowableCol(1)
        req, opt, com = tObj.getArgs()
        dscs = tObj.getArgDescriptions()
        self.topicArgs = {}
        for arg in req + opt:
            lbl = "%s (%s): " % (arg, "required" if arg in req else "optional")
            grid.Add(wx.StaticText(self, label=lbl), 0, wx.ALIGN_RIGHT)
            if useTextCtrl:
                dsc = wx.TextCtrl(self)
                dsc.SetToolTipString(dscs.get(arg, "none"))
                self.topicArgs[arg] = dsc
            else:
                dsc = wx.StaticText(self, label=dscs.get(arg, "None"))
            grid.Add(dsc, 1, wx.ALIGN_LEFT | wx.EXPAND)

        sizer.Add(grid, 1, wx.EXPAND | wx.ALL, 2)
        self.SetSizerAndFit(sizer)
        self.Layout()

        self.Bind(wx.EVT_SIZE, self.OnSize)
        # ~ print dir(tObj)
        ### Some code to explore the types of arguments the listener
        ### expects, strings, booleans, dictionary, etc.

    ##        if tObj.hasListeners():
    ##            listener = tObj.getListeners()[0]
    ##            print listener, listener.__class__, listener.getCallable()
    ##            print inspect.getargspec(listener.getCallable())

    def OnSize(self, evt):
        self.desc.Wrap(evt.GetSize()[0] - 6)

        self.Parent.Refresh()
        evt.Skip()


class MonitorLogger:
    """
    OBSOLETE, needs updating.

    Monitor logger for 'pubsub.*' topics. These topics are used
    automatically in various pubsub calls; e.g. when a new topic is
    created, a 'pubsub.newTopic' message is generated. Note that
    such messages are generated only if pub.setNotificationFlags() was called.

    Each method of DefaultLogger can be given
    to pub.subscribe() as a listener. A method's name indicates
    which 'pubsub' subtopic it can be listening for. E.g. you would
    subscribe the DefaultLogger.subscribe method to listen for
    'pubsub.subscribe' messages:

        from pubsub import pub
        from pubsub import pubsubconf

        from pubsub.utils import DefaultLogger,NotifyByPubsubMessage

        pubsubconf.setNotifier(NotifyByPubsubMessage)
        pub.setNotificationFlags(all=True)
        logger = DefaultLogger(pub)


    Any number of instances can be created. By default, the __init__
    will subscribe instance to all 'pubsub' subtopics. Class can also
    be derived to override default behavior.
    """

    prefix = 'PUBSUB: '
    import sys

    def __init__(self, publisher=None, out=sys.stdout):
        """If publisher is not None, then all self's methods
        are subscribed to the 'pubsub.*' topics by using
        publisher.subscribe(). """
        self._out = out
        self.lastMsg = ''
        self.lastTopic = ''
        self.lastLogTopic = ''
        self.counts = dict(send=0, sub=0, unsub=0, delt=0, newt=0, dead=0, all=0)

        if publisher is not None:
            pub = publisher
            pub.subscribe(self.subscribe, 'pubsub.subscribe')
            pub.subscribe(self.newTopic, 'pubsub.newTopic')
            pub.subscribe(self.delTopic, 'pubsub.delTopic')
            pub.subscribe(self.unsubscribe, 'pubsub.unsubscribe')
            pub.subscribe(self.sendMessage, 'pubsub.sendMessage')
            pub.subscribe(self.deadListener, 'pubsub.deadListener')
            pub.subscribe(self.all, 'pubsub')

    def setOut(self, out):
        self._out = out

    def all(self, msgTopic=pub.AUTO_TOPIC):
        self.counts['all'] += 1
        self.lastLogTopic = msgTopic

    def subscribe(self, topic=None, listener=None, didit=None):
        """Give this to pub.subscribe() as listener of
        'pubsub.subscribe' messages."""
        if didit:
            msg = '%sSubscribed listener %s to topic "%s"\n'
        else:
            msg = '%sSubscription of %s to topic "%s" redundant\n'
        msg = msg % (self.prefix, listener, topic.getName())
        self._out.write(msg)
        self.lastTopic = topic.getName()
        self.lastLogTopic = 'pubsub.subscribe'
        self.counts['sub'] += 1
        self.lastMsg = msg

    ##        print msg

    def unsubscribe(self, topic=None, listener=None, listenerRaw=None):
        """Give this to pub.subscribe() as listener of
        'pubsub.unsubscribe' mesages. """
        msg = '%sUnsubscribed listener %s from topic "%s"\n'
        msg = msg % (self.prefix, listener, topic.getName())
        self._out.write(msg)
        self.lastTopic = topic.getName()
        self.lastLogTopic = 'pubsub.unsubscribe'
        self.lastMsg = msg
        self.counts['unsub'] += 1

    ##        print msg

    def newTopic(self, topic=None, args=None, description=None, required=None):
        """Give this to pub.subscribe() as listener of
        'pubsub.newTopic' messages.
        Messages are only generated when pub.newTopic is called
        explicitly in the code, not on automatic topic generation
        """
        msg = '%sNew topic "%s" created (%s)\n'
        msg = msg % (self.prefix, topic.getName(), description)
        self._out.write(msg)
        self.lastTopic = topic.getName()
        self.lastLogTopic == 'pubsub.newTopic'
        self.lastMsg = msg
        self.counts['newt'] += 1

    ##        print msg

    def delTopic(self, name=None):
        """Give this to pub.subscribe() as listener of
        'pubsub.delTopic' messages. """
        msg = '%sTopic "%s" destroyed\n'
        msg = msg % (self.prefix, name)
        self._out.write(msg)
        self.lastTopic = name
        self.lastLogTopic = 'pubsub.delTopic'
        self.lastMsg = msg
        self.counts['delt'] += 1

    ##        print msg

    def sendMessage(self, topic=None, stage=None):
        """Give this to pub.subscribe() as listener of
        'pubsub.sendMessage' messages. """
        if stage == 'pre':
            msg = '%sSending message of topic "%s"\n'
            msg = msg % (self.prefix, topic.getName())
            self._out.write(msg)
            self.lastTopic = topic.getName()
            self.lastLogTopic = 'pubsub.sendMessage'
            self.lastMsg = msg
            self.counts['send'] += 1
        if stage == 'post':
            msg = "%sSent '%s' message\n"
            msg = msg % (self.prefix, topic.getName())
            self._out.write(msg)
        ##        print msg

    def deadListener(self, topic=None, listener=None):
        """Give this to pub.subscribe() as a listener of
        'deadListener' messages."""
        msg = "%sListener %s for topic '%s' is dead\n"
        msg = msg % (self.prefix, listener, topic.getName())
        self._out.write(msg)
        self.lastMsg = msg
        self.lastTopic = topic.getName()
        self.lastLogTopic = 'pubsub.deadlistener'
        self.counts['dead'] += 1


##        print msg

##class Redirector:
##    """Redirector(wxTextCtrl)
##    Simple wrapper to take output from DefaultLogger (which uses a
##    'write' method to a file-like object) to a wx.TextCtrl (which uses AppendText).
##    """
##    def __init__(self,LogCtrl):
##        self.text = LogCtrl
##        print "Redirector Text",self.text
##        print "Redirector Text parent", self.text.Parent
##
##    def write(self,text):
##        try:
##            self.text.AppendText(text)
##        except:
##            print "cannot print:", text

class LogPanel(wx.Panel):
    def __init__(self, parent):
        wx.Panel.__init__(self, parent, name="log")
        box = wx.StaticBox(self, label="Log")
        sizer = wx.StaticBoxSizer(box, wx.VERTICAL)
        self.Text = wx.TextCtrl(self, style=wx.TE_READONLY | wx.TE_DONTWRAP | wx.TE_MULTILINE)
        sizer.Add(self.Text, 1, wx.EXPAND | wx.GROW | wx.ALL, 2)
        self.SetSizerAndFit(sizer)
        self.Layout()

    def getText(self):
        return self.Text

    def write(self, text):
        ##        print "writing"
        self.Text.AppendText(text)


class LastLogEntryPanel(wx.Panel):
    def __init__(self, parent):
        wx.Panel.__init__(self, parent, name="last log")
        box = wx.StaticBox(self, label="Last Log Entry")

        messageLabel = wx.StaticText(self, label="Message:")
        self.messageText = wx.TextCtrl(self, name="messageText")
        topicLabel = wx.StaticText(self, label="Topic:")
        self.topicText = wx.TextCtrl(self, name="topicText")
        pstopicLabel = wx.StaticText(self, label="psTopic:")
        self.pstopicText = wx.TextCtrl(self, name="pstopicText")

        sizer = wx.StaticBoxSizer(box, wx.VERTICAL)
        row = wx.BoxSizer(wx.HORIZONTAL)
        row.Add(messageLabel, 0, wx.ALIGN_RIGHT | wx.ALIGN_CENTRE_VERTICAL)
        row.Add(self.messageText, 1, wx.LEFT | wx.EXPAND, 3)
        sizer.Add(row, 0, wx.EXPAND)

        row = wx.BoxSizer(wx.HORIZONTAL)
        row.Add(topicLabel, 0, wx.ALIGN_RIGHT | wx.ALIGN_CENTRE_VERTICAL)
        row.Add(self.topicText, 1, wx.LEFT | wx.EXPAND, 3)

        row.Add(pstopicLabel, 0, wx.LEFT | wx.ALIGN_RIGHT | wx.ALIGN_CENTRE_VERTICAL, 6)
        row.Add(self.pstopicText, 1, wx.LEFT | wx.EXPAND, 3)
        sizer.Add(row, 0, wx.EXPAND | wx.TOP, 3)

        self.SetSizerAndFit(sizer)
        self.Layout()

    def SetMessage(self, text):
        self.messageText.SetValue(str(text))

    def SetTopic(self, text):
        self.topicText.SetValue(str(text))

    def SetPSTopic(self, text):
        self.pstopicText.SetValue(str(text))


class CountPanel(wx.Panel):
    def __init__(self, parent):
        wx.Panel.__init__(self, parent, name="counts")
        box = wx.StaticBox(self, label="Counts")
        sizer = wx.StaticBoxSizer(box, wx.VERTICAL)
        grid = wx.GridSizer(cols=7, hgap=2, vgap=2)
        lbls = ['all', 'subscribe', 'unsub', 'send', 'newt', 'delt', 'dead']
        counts = {}
        for c in lbls:
            t = wx.StaticText(self, label="0", name="count_%s" % c,
                              style=wx.ALIGN_CENTRE | wx.ST_NO_AUTORESIZE | wx.SIMPLE_BORDER)
            grid.Add(t, 0,
                     wx.TOP | wx.LEFT | wx.RIGHT | wx.EXPAND | wx.ALIGN_CENTRE_VERTICAL | wx.ALIGN_CENTRE_HORIZONTAL, 2)
        for c in lbls:
            t = wx.StaticText(self, label=c)
            grid.Add(t, 0, wx.ALIGN_CENTRE_VERTICAL | wx.ALIGN_CENTRE_HORIZONTAL)

        sizer.Add(grid, 1, wx.EXPAND | wx.GROW)
        self.SetSizerAndFit(sizer)
        self.Layout()

    def count(self, topicname):
        ##        print "counting:",topicname
        garbage, topicname = topicname.split('.')
        if topicname == 'sendMessage':
            topicname = 'send'
        if topicname == 'unsubscribe':
            topicname = 'unsub'
        if topicname == 'newTopic':
            topicname = 'newt'
        if topicname == 'delTopic':
            topicname = 'delt'
        if topicname == 'deadListener':
            topicname = 'dead'
        lbl = self.FindWindowByName("count_%s" % topicname)
        if lbl:
            c = int(lbl.GetLabel())
            lbl.SetLabel(str(c + 1))
        else:
            print
            "something has gone wrong with", topicname
        alllbl = self.FindWindowByName('count_all')
        alllbl.SetLabel(str(int(alllbl.GetLabel()) + 1))


def test():
    """This test opens the MonitorFrame, as well as a frame
    with a show and hide button
    """

    class showFrame(wx.Frame):
        def __init__(self, parent):
            wx.Frame.__init__(self, parent, title="Monitor Toggle")
            sbtn = wx.Button(self, label="Show Frame")
            sbtn.Bind(wx.EVT_BUTTON, self.SendShowMsg)
            hbtn = wx.Button(self, label="Hide Frame")
            hbtn.Bind(wx.EVT_BUTTON, self.SendHideMsg)

            s = wx.BoxSizer(wx.VERTICAL)
            s.Add(sbtn, 0, wx.EXPAND)
            s.Add(hbtn, 0, wx.EXPAND)
            self.SetSizerAndFit(s)
            self.Layout()

        def SendShowMsg(self, evt):
            print
            "pressed show button"
            pub.sendMessage('monitor.show', show=True)

        def SendHideMsg(self, evt):
            print
            "pressed hide button"
            pub.sendMessage('monitor.hide')

    class myApp(wx.App, MonitorMixin):
        def OnInit(self):
            print
            "Making App"
            MonitorMixin.Init(self)
            show = showFrame(None)
            self.SetTopWindow(show)
            show.Show()
            show.Raise()
            print
            "OnInit end"
            return True

    a = myApp(False)
    print
    "Made app. Starting Mainloop"
    a.MainLoop()
    print
    "Done with main loop"
    a.Destroy()
    pub.delTopic('test')


if __name__ == '__main__':
    print
    "testing"
    test()



"""
:copyright: Copyright since 2006 by Oliver Schoenborn, all rights reserved.
:license: BSD, see LICENSE_BSD_Simple.txt for details.
"""

from typing import List, Mapping

from .listener import Listener
from .topicobj import Topic


class INotificationHandler:
    """
    Defines the interface expected by pubsub for pubsub activity
    notifications. Any instance that supports the same methods, or
    derives from this class, will work as a notification handler
    for pubsub events (see pub.addNotificationHandler).
    """

    def notifySubscribe(self, pubListener: Listener, topicObj: Topic, newSub: bool):
        """
        Called when a listener is subscribed to a topic.
        :param pubListener: the pubsub.core.Listener that wraps subscribed listener.
        :param topicObj: the pubsub.core.Topic object subscribed to.
        :param newSub: false if pubListener was already subscribed.
        """
        raise NotImplementedError

    def notifyUnsubscribe(self, pubListener: Listener, topicObj: Topic):
        """
        Called when a listener is unsubscribed from given topic.
        :param pubListener: the pubsub.core.Listener that wraps unsubscribed listener.
        :param topicObj: the pubsub.core.Topic object unsubscribed from.
        """
        raise NotImplementedError

    def notifyDeadListener(self, pubListener: Listener, topicObj: Topic):
        """
        Called when a listener has been garbage collected.
        :param pubListener: the pubsub.core.Listener that wraps GC'd listener.
        :param topicObj: the pubsub.core.Topic object it was subscribed to.
        """
        raise NotImplementedError

    def notifySend(self, stage: str, topicObj: Topic, pubListener: Listener = None):
        """
        Called multiple times during a sendMessage: once before message
        sending has started (pre), once for each listener about to be sent the
        message, and once after all listeners have received the message (post).

        :param stage: 'pre', 'post', or 'loop'.
        :param topicObj: the Topic object for the message.
        :param pubListener: None for pre and post stages; for loop, the listener
            that is about to be sent the message.
        """
        raise NotImplementedError

    def notifyNewTopic(self, topicObj: Topic, description: str, required: List[str], argsDocs: Mapping[str, str]):
        """
        Called whenever a new topic is added to the topic tree.

        :param topicObj: the Topic object for the message.
        :param description: docstring for the topic.
        :param required: list of message data names (keys in argsDocs) that are required.
        :param argsDocs: dictionary of all message data names, with the
            corresponding docstring.
        """
        raise NotImplementedError

    def notifyDelTopic(self, topicName: str):
        """
        Called whenever a topic is removed from topic tree.
        :param topicName: name of topic removed.
        """
        raise NotImplementedError


class NotificationMgr:
    """
    Manages notifications for tracing pubsub activity. When pubsub takes a
    certain action such as sending a message or creating a topic, and
    the notification flag for that activity is True, all registered
    notification handlers get corresponding method called with information
    about the activity, such as which listener subscribed to which topic.
    See INotificationHandler for which method gets called for each activity.

    If more than one notification handler has been registered, the order in
    which they are notified is unspecified (do not rely on it).

    Note that this manager automatically unregisters all handlers when
    the Python interpreter exits, to help avoid NoneType exceptions during
    shutdown. This "shutdown" starts when the last line of app "main" has
    executed; the Python interpreter then starts cleaning up, garbage
    collecting everything, which could lead to various pubsub notifications
    -- by then they should be of no interest -- such as dead
    listeners, etc.
    """

    def __init__(self, notificationHandler: INotificationHandler = None):
        self.__notifyOnSend = False
        self.__notifyOnSubscribe = False
        self.__notifyOnUnsubscribe = False

        self.__notifyOnNewTopic = False
        self.__notifyOnDelTopic = False
        self.__notifyOnDeadListener = False

        self.__handlers = []
        if notificationHandler is not None:
            self.addHandler(notificationHandler)

        self.__atExitRegistered = False

    def addHandler(self, handler: INotificationHandler):
        if not self.__atExitRegistered:
            self.__registerForAppExit()
        self.__handlers.append(handler)

    def getHandlers(self) -> List[INotificationHandler]:
        return self.__handlers[:]

    def clearHandlers(self):
        self.__handlers = []

    def notifySubscribe(self, *args, **kwargs):
        if self.__notifyOnSubscribe and self.__handlers:
            for handler in self.__handlers:
                handler.notifySubscribe(*args, **kwargs)

    def notifyUnsubscribe(self, *args, **kwargs):
        if self.__notifyOnUnsubscribe and self.__handlers:
            for handler in self.__handlers:
                handler.notifyUnsubscribe(*args, **kwargs)

    def notifySend(self, *args, **kwargs):
        if self.__notifyOnSend and self.__handlers:
            for handler in self.__handlers:
                handler.notifySend(*args, **kwargs)

    def notifyNewTopic(self, *args, **kwargs):
        if self.__notifyOnNewTopic and self.__handlers:
            for handler in self.__handlers:
                handler.notifyNewTopic(*args, **kwargs)

    def notifyDelTopic(self, *args, **kwargs):
        if self.__notifyOnDelTopic and self.__handlers:
            for handler in self.__handlers:
                handler.notifyDelTopic(*args, **kwargs)

    def notifyDeadListener(self, *args, **kwargs):
        if self.__notifyOnDeadListener and self.__handlers:
            for handler in self.__handlers:
                handler.notifyDeadListener(*args, **kwargs)

    def getFlagStates(self) -> Mapping[str, bool]:
        """Return state of each notification flag, as a dict."""
        return dict(
            subscribe=self.__notifyOnSubscribe,
            unsubscribe=self.__notifyOnUnsubscribe,
            deadListener=self.__notifyOnDeadListener,
            sendMessage=self.__notifyOnSend,
            newTopic=self.__notifyOnNewTopic,
            delTopic=self.__notifyOnDelTopic,
        )

    def setFlagStates(self, subscribe: bool = None, unsubscribe: bool = None, deadListener: bool = None,
                      sendMessage: bool = None, newTopic: bool = None, delTopic: bool = None, all: bool = None):
        """
        Set the notification flag on/off for various aspects of pubsub.
        The kwargs that are None are left at their current value. The 'all',
        if not None, is set first. E.g.

            mgr.setFlagStates(all=True, delTopic=False)

        will toggle all notifications on, but will turn off the 'delTopic'
        notification.
        """
        if all is not None:
            # ignore all other arg settings, and set all of them to true:
            numArgs = 7  # how many args in this method
            self.setFlagStates(all=None, *((numArgs - 1) * [all]))

        if sendMessage is not None:
            self.__notifyOnSend = sendMessage
        if subscribe is not None:
            self.__notifyOnSubscribe = subscribe
        if unsubscribe is not None:
            self.__notifyOnUnsubscribe = unsubscribe

        if newTopic is not None:
            self.__notifyOnNewTopic = newTopic
        if delTopic is not None:
            self.__notifyOnDelTopic = delTopic
        if deadListener is not None:
            self.__notifyOnDeadListener = deadListener

    def __registerForAppExit(self):
        import atexit
        atexit.register(self.clearHandlers)
        self.__atExitRegistered = True



"""
Miscellaneous utility items

:copyright: Copyright since 2006 by Oliver Schoenborn, all rights reserved.
:license: BSD, see LICENSE_BSD_Simple.txt for details.
"""

from typing import Any


def annotationType(sig_obj: Any):
    """
    Use to forward declare an annotation type, i.e. a type that will be fully defined later but is not available
    at the annotation location in the source code. The following example shows a MyType class with a method that
    accepts a MyType instance; annotating this cannot be done in Python::

        class MyType:
            def copy_from(other: MyType):  # interpreter will croak
                ...

    The recommended approach is to use a string annotation, but this is rather unnatural in Python code::

        class MyType:
            def copy_from(other: 'MyType'):
                ...

    The annotationType function allows a more pythonic syntax:

        @annotationType
        class MyType:
            pass

        class MyType:
            def copy_from(other: MyType):
                ...

    This decorator doesn't actually do anything to its argument.
    """
    return sig_obj
